//$$ boolean.hxx                     Boolean class

#ifndef BOOL_LIB
#define BOOL_LIB 0

class BOOL
{
   int value;
public:
   BOOL(int b) { value = b ? 1 : 0; }
   BOOL(void* b) { value = b ? 1 : 0; }
   BOOL() {}
   operator int() const { return value; }
   BOOL operator&&(const BOOL& b) const { return value && b.value; }
   BOOL operator||(const BOOL& b) const { return value || b.value; }
};

#define FALSE 0
#define TRUE 1

#endif
