//$$ newmatio.hxx         definition file for matrix package input/output

// Copyright (C) 1991: R B Davies and DSIR

#ifndef MATRIXIO_LIB
#define MATRIXIO_LIB 0

#include "newmat.hxx"

/**************************** input/output *****************************/

ostream& operator<<(ostream&, BaseMatrix&);

ostream& operator<<(ostream&, const GeneralMatrix&);


#endif
